﻿// ClickAndDeformText (c) 2009 by Charles Petzold
using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using System.Windows.Media;

namespace ClickAndDeformText
{
    public partial class MainPage : UserControl
    {
        PathGeometry basePathGeometry;
        Rect boundingRect;
        bool isManipulating;

        public MainPage()
        {
            InitializeComponent();

            // Clone the PathGeometry and get its bounding box
            basePathGeometry = CloneFlattenedPathGeometry(textPathGeometry, ref boundingRect);

            // Initialize the thumb positions
            Canvas.SetLeft(thumbUpper0, boundingRect.Left);
            Canvas.SetLeft(thumbLower0, boundingRect.Left);
            Canvas.SetLeft(thumbUpper1, boundingRect.Left + boundingRect.Width / 3);
            Canvas.SetLeft(thumbLower1, boundingRect.Left + boundingRect.Width / 3);
            Canvas.SetLeft(thumbUpper2, boundingRect.Left + 2 * boundingRect.Width / 3);
            Canvas.SetLeft(thumbLower2, boundingRect.Left + 2 * boundingRect.Width / 3);
            Canvas.SetLeft(thumbUpper3, boundingRect.Right);
            Canvas.SetLeft(thumbLower3, boundingRect.Right);

            Canvas.SetTop(thumbUpper0, boundingRect.Top);
            Canvas.SetTop(thumbUpper1, boundingRect.Top);
            Canvas.SetTop(thumbUpper2, boundingRect.Top);
            Canvas.SetTop(thumbUpper3, boundingRect.Top);

            Canvas.SetTop(thumbLower0, boundingRect.Bottom);
            Canvas.SetTop(thumbLower1, boundingRect.Bottom);
            Canvas.SetTop(thumbLower2, boundingRect.Bottom);
            Canvas.SetTop(thumbLower3, boundingRect.Bottom);

            // Sync the Bezier curves with the thumbs
            SyncThumbsAndBeziers();
        }

        // Toggle between manipulating an not manipulating
        protected override void OnMouseLeftButtonDown(MouseButtonEventArgs args)
        {
            if (!isManipulating)
            {
                textPath.Opacity = 0.25;
                manipulateCanvas.Visibility = Visibility.Visible;
                isManipulating = true;
            }
            else
            {
                textPath.Opacity = 1;
                manipulateCanvas.Visibility = Visibility.Collapsed;
                isManipulating = false;
            }
        }

        // Move the thumbs and do everything else
        void OnThumbDragDelta(object sender, DragDeltaEventArgs args)
        {
            Thumb thumb = sender as Thumb;
            Canvas.SetLeft(thumb, Canvas.GetLeft(thumb) + args.HorizontalChange);
            Canvas.SetTop(thumb, Canvas.GetTop(thumb) + args.VerticalChange);
            SyncThumbsAndBeziers();
            DeformText();
        }

        // Set the Bezier points to the thumb locations
        void SyncThumbsAndBeziers()
        {
            upperBezierFigure.StartPoint = new Point(Canvas.GetLeft(thumbUpper0), Canvas.GetTop(thumbUpper0));
            upperBezierSegment.Point1    = new Point(Canvas.GetLeft(thumbUpper1), Canvas.GetTop(thumbUpper1));
            upperBezierSegment.Point2    = new Point(Canvas.GetLeft(thumbUpper2), Canvas.GetTop(thumbUpper2));
            upperBezierSegment.Point3    = new Point(Canvas.GetLeft(thumbUpper3), Canvas.GetTop(thumbUpper3));

            lowerBezierFigure.StartPoint = new Point(Canvas.GetLeft(thumbLower0), Canvas.GetTop(thumbLower0));
            lowerBezierSegment.Point1    = new Point(Canvas.GetLeft(thumbLower1), Canvas.GetTop(thumbLower1));
            lowerBezierSegment.Point2    = new Point(Canvas.GetLeft(thumbLower2), Canvas.GetTop(thumbLower2));
            lowerBezierSegment.Point3    = new Point(Canvas.GetLeft(thumbLower3), Canvas.GetTop(thumbLower3));

            // Tried to use ElementName binding in XAML for this stuff, but got the informative AG_E_PARSER_BAD_PROPERTY_VALUE error
            upperBezierLine1.StartPoint = upperBezierFigure.StartPoint;
            upperBezierLine1.EndPoint = upperBezierSegment.Point1;
            upperBezierLine2.StartPoint = upperBezierSegment.Point3;
            upperBezierLine2.EndPoint = upperBezierSegment.Point2;

            lowerBezierLine1.StartPoint = lowerBezierFigure.StartPoint;
            lowerBezierLine1.EndPoint = lowerBezierSegment.Point1;
            lowerBezierLine2.StartPoint = lowerBezierSegment.Point3;
            lowerBezierLine2.EndPoint = lowerBezierSegment.Point2;
        }

        // Set textPathGeometry from basePathGeometry but deformed based on Bezier curves
        void DeformText()
        {
            for (int figureIndex = 0; figureIndex < textPathGeometry.Figures.Count; figureIndex++)
            {
                PathFigure textPathFigure = textPathGeometry.Figures[figureIndex];
                PathFigure basePathFigure = basePathGeometry.Figures[figureIndex];
                textPathFigure.StartPoint = DeformPoint(basePathFigure.StartPoint);

                for (int segmentIndex = 0; segmentIndex < textPathFigure.Segments.Count; segmentIndex++)
                {
                    PolyLineSegment textPolyLineSegment = textPathFigure.Segments[segmentIndex] as PolyLineSegment;
                    PolyLineSegment basePolyLineSegment = basePathFigure.Segments[segmentIndex] as PolyLineSegment;

                    for (int pointIndex = 0; pointIndex < textPolyLineSegment.Points.Count; pointIndex++)
                    {
                        textPolyLineSegment.Points[pointIndex] = DeformPoint(basePolyLineSegment.Points[pointIndex]);
                    }
                }
            }
        }

        // xFraction corresonds to t in the Bezier parametric formulas; yFraction does a weighted average
        Point DeformPoint(Point basePoint)
        {
            double xFraction = (basePoint.X - boundingRect.Left) / boundingRect.Width;
            double yFraction = (basePoint.Y - boundingRect.Top) / boundingRect.Height;

            Point upperPoint = GetBezierPoint(upperBezierFigure.StartPoint,
                                              upperBezierSegment.Point1,
                                              upperBezierSegment.Point2,
                                              upperBezierSegment.Point3,
                                              xFraction);

            Point lowerPoint = GetBezierPoint(lowerBezierFigure.StartPoint,
                                              lowerBezierSegment.Point1,
                                              lowerBezierSegment.Point2,
                                              lowerBezierSegment.Point3,
                                              xFraction);

            return new Point(upperPoint.X * (1 - yFraction) + lowerPoint.X * yFraction,
                             upperPoint.Y * (1 - yFraction) + lowerPoint.Y * yFraction);
        }

        Point GetBezierPoint(Point p0, Point p1, Point p2, Point p3, double t)
        {
            double x = (1 - t) * (1 - t) * (1 - t) * p0.X + 3 * t * (1 - t) * (1 - t) * p1.X + 3 * t * t * (1 - t) * p2.X + t * t * t * p3.X ;
            double y = (1 - t) * (1 - t) * (1 - t) * p0.Y + 3 * t * (1 - t) * (1 - t) * p1.Y + 3 * t * t * (1 - t) * p2.Y + t * t * t * p3.Y;
            return new Point(x, y);
        }

        static PathGeometry CloneFlattenedPathGeometry(PathGeometry pathGeometry, ref Rect boundingRect)
        {
            double xMin = Double.MaxValue;
            double yMin = Double.MaxValue;
            double xMax = Double.MinValue;
            double yMax = Double.MinValue;

            PathGeometry clonedPathGeometry = new PathGeometry()
            {
                FillRule = pathGeometry.FillRule
            };

            foreach (PathFigure pathFigure in pathGeometry.Figures)
            {
                PathFigure clonedPathFigure = new PathFigure()
                {
                    IsClosed = pathFigure.IsClosed,
                    IsFilled = pathFigure.IsFilled,
                    StartPoint = pathFigure.StartPoint
                };

                NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, clonedPathFigure.StartPoint);

                foreach (PathSegment pathSegment in pathFigure.Segments)
                {
                    PolyLineSegment polyLineSegment = pathSegment as PolyLineSegment;

                    if (polyLineSegment == null)
                        throw new NotImplementedException("CloneFlattenedPathGeometry does not support PathSegment " +
                                                          "objects of type " + pathSegment.GetType().Name);

                    PolyLineSegment clonedPolyLineSegment = new PolyLineSegment();

                    foreach (Point point in polyLineSegment.Points)
                    {
                        NewMinMax(ref xMin, ref xMax, ref yMin, ref yMax, point);
                        clonedPolyLineSegment.Points.Add(point);
                    }
                    clonedPathFigure.Segments.Add(clonedPolyLineSegment);
                }
                clonedPathGeometry.Figures.Add(clonedPathFigure);
            }
            boundingRect = new Rect(new Point(xMin, yMin), new Point(xMax, yMax));
            return clonedPathGeometry;
        }

        static void NewMinMax(ref double xMin, ref double xMax, ref double yMin, ref double yMax, Point pt)
        {
            xMin = Math.Min(xMin, pt.X);
            xMax = Math.Max(xMax, pt.X);
            yMin = Math.Min(yMin, pt.Y);
            yMax = Math.Max(yMax, pt.Y);
        }
    }
}
